theory Dependency_Graph
imports "Q-Restricted_Rewriting" Graph Name Substitution State_Monad
begin

section {* Preliminaries *}

type_synonym ('f, 'v) dg = "('f, 'v) term graph"
type_synonym ('f, 'v) terms = "('f, 'v) term set"

text {* Combined set of left-hand sides and right-hand sides of a TRS. *}
definition
  terms :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term set"
where
  "terms R \<equiv> lhss R \<union> rhss R"

definition
  DG :: "('f, 'v) trs \<Rightarrow> ('f, 'v) terms \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) dg"
where
  "DG P Q R \<equiv> (terms P, {(t, u) | s t u v. (s, t) \<in> P
    \<and> (u, v) \<in> P \<and> (\<exists>\<sigma> \<tau>. (t \<cdot> \<sigma>, u \<cdot> \<tau>) \<in> (qrstep Q R)^*
    \<and> s \<cdot> \<sigma> \<in> NF_terms Q \<and> u \<cdot> \<tau> \<in> NF_terms Q)})"

section {* A semantic version of cap *}

text {* Applying a substitution to every term of a given set. *}
abbreviation
  subst_apply_set :: "('f, 'v) terms \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) terms" (infixl "\<cdot>set" 60)
where
  "T \<cdot>set \<sigma> \<equiv> (\<lambda>t. t \<cdot> \<sigma>) ` T"

text {*
The \emph{cap} of a term is the `upper' part which is definitely not
changed by rewriting. Parts which are potentially changed, are replaced by
fresh variables (represented by @{term "Var (Inl ())"} below).
*}
fun
  cap :: "('f, 'v) trs \<Rightarrow> ('f, 'v) terms \<Rightarrow> ('f, 'v) terms
    \<Rightarrow> ('f, 'v) term  \<Rightarrow> ('f, unit + 'v) term"
where
  "cap R Q S (Var x) = (if (\<exists>\<sigma>. S \<cdot>set \<sigma> \<subseteq> NF_terms Q \<and> Var x \<cdot> \<sigma> \<notin> NF (qrstep Q R))
    then Var (Inl ())
    else Var (Inr x))" |
  "cap R Q S (Fun f ts) = (
    if (\<exists>\<sigma> u. S \<cdot>set \<sigma> \<subseteq> NF_terms Q \<and> (Fun f ts \<cdot> \<sigma>, u) \<in> (qrstep Q R)^* O rqrstep Q R)
      then Var (Inl ())
      else Fun f (map (cap R Q S) ts))"

declare size_simps[simp]
fun
  fresh_substs_class :: "('f, unit + 'v) term \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) terms" ("[_]'__")
where
  "[Var (Inl ())]_\<sigma> = UNIV" |
  "[Var (Inr x) ]_\<sigma> = {Var x \<cdot> \<sigma>}" |
  "[Fun f ss    ]_\<sigma> = {Fun f ts | ts.
    length ts = length ss \<and> (\<forall>i<length ss. (ts!i) \<in> [ss!i]_\<sigma>)}"
declare size_simps[simp del]

abbreviation
  fresh_class :: "('f, unit + 'v) term \<Rightarrow> ('f, 'v) terms" ("[_]")
where
  "[t] \<equiv> [t]_\<iota>"

lemma fresh_substs_class_not_empty[simp]:
  "[t]_\<sigma> \<noteq> {}"
proof (induct t)
  case (Var x) thus ?case by (cases x) auto
next
  case (Fun f ts)
  hence "\<forall>t\<in>set ts. \<exists>s. s \<in> [t]_\<sigma>" by auto
  from bchoice[OF this] obtain s
    where in_class: "\<forall>t\<in>set ts. s t \<in> [t]_\<sigma>" by auto
  let ?ss = "map (\<lambda>i. (s (ts!i))) [0..<length ts]"
  have len: "length ?ss = length ts" by simp
  from in_class have "\<forall>i<length ts. (s (ts!i)) \<in> [ts!i]_\<sigma>" by auto
  hence "\<forall>i<length ts. (?ss!i) \<in> [ts!i]_\<sigma>" by simp
  with len have "Fun f ?ss \<in> [Fun f ts]_\<sigma>" by auto
  thus ?case by best
qed

lemma fresh_substs_class_subst: "[t]_\<tau> \<cdot>set \<sigma> \<subseteq> [t]_(\<tau> \<circ>s \<sigma>)"
proof (induct t arbitrary: \<sigma> \<tau>)
  case (Var x)
  show ?case
  proof (cases x)
    case (Inl n)
    have "[Var x]_\<tau> \<cdot>set \<sigma> = UNIV \<cdot>set \<sigma>" by (simp add: Inl)
    moreover have "[Var x]_(\<tau> \<circ>s \<sigma>) = UNIV" by (simp add: Inl)
    moreover have "UNIV \<cdot>set \<sigma> \<subseteq> UNIV" by auto
    ultimately show ?thesis by blast
  next
    case (Inr y) thus ?thesis by simp
  qed
next
  case (Fun f ts)
  show ?case
  proof
    fix t assume in_Fun: "t \<in> [Fun f ts]_\<tau> \<cdot>set \<sigma>"
    then obtain ss where len: "length ss = length ts" and "\<forall>i<length ts. (ss!i) \<in> [ts!i]_\<tau>"
      and t: "t = Fun f (map (\<lambda>t. t\<cdot>\<sigma>) ss)" by auto
    moreover from Fun have "\<forall>i<length ts. [ts!i]_\<tau> \<cdot>set \<sigma> \<subseteq> [ts!i]_(\<tau> \<circ>s \<sigma>)" by auto
    ultimately have "\<forall>i<length ts. (ss!i) \<cdot> \<sigma> \<in> [ts!i]_(\<tau> \<circ>s \<sigma>)" by auto
    with len show "t \<in> [Fun f ts]_(\<tau> \<circ>s \<sigma>)" unfolding t by auto
  qed
qed

lemma fresh_class_subst: "[t] \<cdot>set \<sigma> \<subseteq> [t]_\<sigma>"
proof -
  have "[t]_\<iota> \<cdot>set \<sigma> \<subseteq> [t]_(\<iota> \<circ>s \<sigma>)" by (rule fresh_substs_class_subst)
  thus ?thesis by simp
qed

definition
  is_ecap ::
    "(('f, 'v) trs \<Rightarrow> ('f, 'v) terms \<Rightarrow> ('f, 'v) terms \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, unit + 'v) term)
      \<Rightarrow> bool"
where
  "is_ecap ecap \<equiv> \<forall>R Q S t f ts.
    [cap R Q S t] \<subseteq> [ecap R Q S t]
    \<and> ecap R Q S (Fun f ts) \<in> {Var (Inl ()), Fun f (map (ecap R Q S) ts)}"

lemma is_ecapI[intro]:
  assumes "\<And>R Q S t f ts. [cap R Q S t] \<subseteq> [ecap R Q S t]
    \<and> ecap R Q S (Fun f ts) \<in> {Var (Inl ()), Fun f (map (ecap R Q S) ts)}"
  shows "is_ecap ecap"
using assms unfolding is_ecap_def by auto

lemma is_ecapE[elim]:
  assumes "is_ecap ecap"
    and "\<lbrakk>[cap R Q S t] \<subseteq> [ecap R Q S t];
      ecap R Q S (Fun f ts) \<in> {Var (Inl ()), Fun f (map (ecap R Q S) ts)}\<rbrakk> \<Longrightarrow> P"
  shows P
using assms unfolding is_ecap_def by simp

lemma is_ecap_cap[intro]: "is_ecap cap"
proof
  fix R Q S t f ts
    show "[cap R Q S t] \<subseteq> [cap R Q S t]
      \<and> cap R Q S (Fun f ts) \<in> {Var (Inl ()), Fun f (map (cap R Q S) ts)}"
  by (induct t) simp_all
qed

lemma UNIV_term_not_singleton[intro]: "(UNIV::('f,'v) terms) = {t} \<Longrightarrow> False"
by (induct t) auto

lemma UNIV_fresh_substs_classE:
  assumes "UNIV = [t]_\<sigma>" and "t = Var (Inl ()) \<Longrightarrow> P" shows P
proof (cases t)
  case (Var x) show ?thesis using assms unfolding Var by (cases x) auto
next
  case (Fun f ts) show ?thesis using assms unfolding Fun by auto
qed

lemma Var_Inr_fresh_classE:
  assumes "[Var (Inr x)] \<subseteq> [t]" and "t = Var (Inl ()) \<Longrightarrow> P"
    and "t = Var (Inr x) \<Longrightarrow> P"
  shows P
proof (cases t)
  case (Var y) show ?thesis using assms unfolding Var by (cases y) auto
next
  case (Fun f ts) show ?thesis using assms unfolding Fun by auto
qed

lemma Fun_fresh_classE:
  assumes "Fun f ts \<in> [t]"
    and "t = Var (Inl ()) \<Longrightarrow> P"
    and "\<And>ss. \<lbrakk>t = Fun f ss;length ts = length ss; \<forall>i<length ss. ts!i \<in> [ss!i]\<rbrakk> \<Longrightarrow> P"
  shows P
proof (cases t)
  case (Var x) show ?thesis using assms unfolding Var by (cases x) auto
next
  case (Fun g ss) show ?thesis using assms unfolding Fun by auto
qed

lemma fresh_class_Var_fresh: "t \<in> [Var (Inl ())]_\<sigma>" by simp

lemma fresh_class_Var_old:
  assumes "t \<in> [Var (Inr x)]_\<sigma>" and "t = Var x \<cdot> \<sigma> \<Longrightarrow> P" shows P
using assms by simp

lemma equiv_class_Fun:
  assumes "t \<in> [Fun f ts]_\<sigma>"
    and "\<And>ss. \<lbrakk>t = Fun f ss; length ss = length ts; \<forall>i<length ts. ss!i \<in> [ts!i]_\<sigma>\<rbrakk> \<Longrightarrow> P"
  shows P
using assms by auto

lemma cap_Var_fresh:
  assumes "is_ecap ecap" and "cap R Q S (Var x) = Var (Inl ())"
  shows "ecap R Q S (Var x) = Var (Inl ())"
using assms(1) proof
  assume cap: "[cap R Q S (Var x)] \<subseteq> [ecap R Q S (Var x)]"
  show "ecap R Q S (Var x) = Var (Inl ())"
  proof (cases "cap R Q S (Var x) = Var (Inl ())")
    case True
    from cap[unfolded True] show ?thesis using UNIV_fresh_substs_classE[of _ \<iota>] by auto
  next
    case False
    hence "cap R Q S (Var x) = Var (Inr x)" by (auto simp: split_ifs)
    with assms have False by auto
    thus ?thesis by simp
  qed
qed

lemma is_ecap_Var:
  assumes "is_ecap ecap" and "ecap R Q S (Var x) = Var (Inr x)"
  shows "cap R Q S (Var x) = Var (Inr x)"
proof -
  from assms have "ecap R Q S (Var x) \<noteq> Var (Inl ())" by auto
  with cap_Var_fresh[OF assms(1)]
    have "cap R Q S (Var x) \<noteq> Var (Inl ())" by blast
  thus ?thesis by (simp add: split_ifs)
qed

lemma cap_Var_old:
  assumes "is_ecap ecap" and "cap R Q S (Var x) = Var (Inr x)"
  shows "ecap R Q S (Var x) = Var (Inl ()) \<or> ecap R Q S (Var x) = Var (Inr x)"
using assms(1) proof
  assume cap: "[cap R Q S (Var x)] \<subseteq> [ecap R Q S (Var x)]"
  from Var_Inr_fresh_classE[OF this[unfolded assms]] show ?thesis by auto
qed

lemma cap_Var_cases:
  assumes "cap R Q S (Var x) = Var (Inr x) \<Longrightarrow> P"
    and "cap R Q S (Var x) = Var (Inl ()) \<Longrightarrow> P"
  shows P
using assms by (auto simp: split_ifs)

lemma ecap_Var_not_fresh:
  assumes "is_ecap ecap" and "ecap R Q S (Var x) \<noteq> Var (Inl ())"
  shows "ecap R Q S (Var x) = Var (Inr x)"
using assms(1) proof
  assume cap: "[cap R Q S (Var x)] \<subseteq> [ecap R Q S (Var x)]"
  show ?thesis
  proof (rule cap_Var_cases)
    assume "cap R Q S (Var x) = Var (Inr x)"
    from cap_Var_old[OF assms(1) this] and assms show ?thesis by simp
  next
    assume "cap R Q S (Var x) = Var (Inl ())"
    from cap_Var_fresh[OF assms(1) this] and assms show ?thesis by simp
  qed
qed

lemma cap_Fun_cases:
  assumes "cap R Q S (Fun f ts) = Var (Inl ()) \<Longrightarrow> P"
    and "cap R Q S (Fun f ts) = Fun f (map (cap R Q S) ts) \<Longrightarrow> P"
  shows P
using assms unfolding cap.simps split_ifs by blast

lemma cap_Fun_fresh:
  assumes "is_ecap ecap" and "cap R Q S (Fun f ts) = Var (Inl ())"
  shows "ecap R Q S (Fun f ts) = Var (Inl ())"
using assms(1) proof
  assume "[cap R Q S (Fun f ts)] \<subseteq> [ecap R Q S (Fun f ts)]"
  from UNIV_fresh_substs_classE[of _ \<iota>] and this[unfolded assms] show ?thesis by auto
qed

lemma ecap_Fun:
  assumes "is_ecap ecap"
  shows "ecap R Q S (Fun f ts) = Var (Inl ())
    \<or> ecap R Q S (Fun f ts) = Fun f (map (ecap R Q S) ts)"
using assms by auto


(* Lemma 3.8 of Rene's Thesis *)
lemma
  assumes ecap: "is_ecap ecap"
    and step: "(t \<cdot> \<sigma>, u) \<in> (qrstep Q R)^*"
    and NF: "S \<cdot>set \<sigma> \<subseteq> NF_terms Q"
  shows "u \<in> [ecap R Q S t]_\<sigma>"
using step proof (induct t arbitrary: u)
  case (Var x)
  show ?case
  proof (cases "Var x \<cdot> \<sigma> \<in> NF (qrstep Q R)")
    case True
    show ?thesis
    proof (cases rule: cap_Var_cases)
      assume "cap R Q S (Var x) = Var (Inl ())"
      from cap_Var_fresh[OF ecap this] show ?thesis by auto
    next
      assume "cap R Q S (Var x) = Var (Inr x)"
      from cap_Var_old[OF ecap this]
        have "ecap R Q S (Var x) = Var (Inl ()) \<or> ecap R Q S (Var x) = Var (Inr x)" .
      thus ?thesis
      proof
        assume "ecap R Q S (Var x) = Var (Inl ())"
        thus ?thesis by auto
      next
        assume "ecap R Q S (Var x) = Var (Inr x)"
        moreover from Var and True have "u = Var x \<cdot> \<sigma>" by (induct rule: rtrancl.induct) auto
        ultimately show ?thesis by auto
      qed
    qed
  next
    case False
    with NF have "cap R Q S (Var x) = Var (Inl ())" by auto
    from cap_Var_fresh[OF ecap this] show ?thesis by auto
  qed
next
  case (Fun f ss)
  have "(\<exists>us. length us = length ss
    \<and> u = Fun f us \<and> (\<forall>i<length ss. ((ss ! i) \<cdot> \<sigma>, us ! i) \<in> (qrstep Q R)^*))
    \<or> (Fun f ss \<cdot> \<sigma>, u) \<in> (qrstep Q R)^* O (rqrstep Q R) O (qrstep Q R)^*"
    using qrsteps_rqrstep_cases[of f "map (\<lambda>t. t\<cdot>\<sigma>) ss" u Q R,
      unfolded subst_apply_term.simps[symmetric], OF Fun(2), unfolded length_map]
      using nth_map[of _ ss "\<lambda>t. t \<cdot> \<sigma>"] by auto
  thus ?case
  proof
    assume "(Fun f ss \<cdot> \<sigma>, u) \<in> (qrstep Q R)^* O (rqrstep Q R) O (qrstep Q R)^*"
    then obtain v where "(Fun f ss \<cdot> \<sigma>, v) \<in> (qrstep Q R)^* O (rqrstep Q R)" by auto
    with NF
    have "cap R Q S (Fun f ss) = Var (Inl ())" by auto
    from cap_Fun_fresh[OF ecap this] show ?thesis by auto
  next
    assume "\<exists>us. length us = length ss
      \<and> u = Fun f us \<and> (\<forall>i<length ss. ((ss ! i) \<cdot> \<sigma>, us ! i) \<in> (qrstep Q R)^*)"
    then obtain us where len: "length us = length ss" and u: "u = Fun f us"
      and steps: "\<forall>i<length ss. ((ss!i)\<cdot>\<sigma>, us!i) \<in> (qrstep Q R)^*" by auto
    with Fun have args: "\<forall>i<length ss. us!i \<in> [ecap R Q S (ss!i)]_\<sigma>" by auto
    from ecap_Fun[OF ecap]
      have "ecap R Q S (Fun f ss) = Var (Inl ())
        \<or> ecap R Q S (Fun f ss) = Fun f (map (ecap R Q S) ss)" .
    thus ?thesis
    proof
      assume "ecap R Q S (Fun f ss) = Var (Inl ())" thus ?thesis by auto
    next
      assume ecap_map: "ecap R Q S (Fun f ss) = Fun f (map (ecap R Q S) ss)"
      from len and args have "Fun f us \<in> [ecap R Q S (Fun f ss)]_\<sigma>"
        unfolding ecap_map by auto
      thus ?thesis unfolding u .
    qed
  qed
qed

fun
  freshify :: "('f, unit + 'v) term \<Rightarrow> (('f, nat + 'v) term, nat) sm"
where
  "freshify (Var (Inr x))  = return (Var (Inr x))" |
  "freshify (Var (Inl ())) = do { i \<leftarrow> update Suc; return (Var (Inl i)) }" |
  "freshify (Fun f ts) = do {
    ss \<leftarrow> mapm freshify ts;
    return (Fun f ss)
  }"

abbreviation
  ren :: "nat \<Rightarrow> ('f, unit + 'v) term \<Rightarrow> ('f, nat + 'v) term"
where
  "ren n t \<equiv> fst (run (freshify t) n)"

abbreviation
  right_vars :: "('f, 'v + 'w) term \<Rightarrow> 'w set"
where
  "right_vars t \<equiv> {y | x y. x \<in> vars_term t \<and> x = Inr y}"

abbreviation
  left_vars :: "('f, 'v + 'w) term \<Rightarrow> 'v set"
where
  "left_vars t \<equiv> {y | x y. x \<in> vars_term t \<and> x = Inl y}"

fun
  fresh_subst :: "(nat \<Rightarrow> ('f, 'v) term) \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, nat + 'v) subst"
where
  "fresh_subst f (Subst s) = Subst (\<lambda>v.
    case v of Inl n \<Rightarrow> map_vars Inr (f n)
            | Inr x \<Rightarrow> map_vars Inr (s x))"

end



