theory State_Monad
imports "~~/src/HOL/Library/Monad_Syntax"
begin

type_synonym
  ('a, 's) sm = "'s \<Rightarrow> 'a \<times> 's"

definition
  bind :: "('a, 's) sm \<Rightarrow> ('a \<Rightarrow> ('b, 's) sm) \<Rightarrow> ('b, 's) sm"
where
  "bind m n s \<equiv> (
    let (x, s') = m s
    in n x s'
  )"

setup {*
  Adhoc_Overloading.add_variant @{const_name Monad_Syntax.bind} @{const_name bind}
*}

abbreviation (input)
  return :: "'a \<Rightarrow> ('a, 's) sm"
where
  "return \<equiv> Pair"

definition
  run :: "('a, 's) sm \<Rightarrow> 's \<Rightarrow> 'a \<times> 's"
where
  "run m s \<equiv> m s"

subsection {* Monad Laws *}

lemma return_bind: "(Pair x \<guillemotright>= f) = f x"
by (rule ext) (simp add: bind_def)

lemma bind_return: "(m \<guillemotright>= Pair) = m" by (rule ext) (simp add: bind_def)

lemma bind_assoc: "(((m :: ('a, 's) sm) \<guillemotright>= f) \<guillemotright>= g) = (m \<guillemotright>= (\<lambda>x. f x \<guillemotright>= g))"
by (rule ext) (simp add: bind_def, auto simp: split_def)

definition
  update :: "('s \<Rightarrow> 's) \<Rightarrow> ('s, 's) sm"
where
  "update f s \<equiv> (s, f s)"

lemma run_bind:
  "run (m \<guillemotright>= f) s = split f (run m s)"
by (simp add: run_def split_def bind_def Let_def)

lemma run_return: "run (Pair x) y = (x, y)" by (simp add: run_def)

lemma run_update: "run (update f) s = (s, f s)"
unfolding run_def update_def ..

lemmas run_simps = run_bind run_return run_update

fun
  foldm :: "('a \<Rightarrow> 'b \<Rightarrow> ('a, 's) sm) \<Rightarrow> 'a \<Rightarrow> 'b list \<Rightarrow> ('a, 's) sm"
where 
  "foldm f d [] = return d" |
  "foldm f d (x # xs) = do { y \<leftarrow> f d x; foldm f y xs }"

lemma foldm_cong[fundef_cong]:
  assumes "\<And>x y. x \<in> set xs \<Longrightarrow> f y x = g y x"
  shows "foldm f d xs = foldm g d xs"
using assms by (induct xs arbitrary: d ) (simp_all)

fun
  sequence :: "('a, 's) sm list \<Rightarrow> ('a list, 's) sm"
where
  "sequence [] = Pair []" |
  "sequence (m # ms) = do {
    x \<leftarrow> m;
    xs \<leftarrow> sequence ms;
    return (x # xs)
  }"

definition
  mapm :: "('a \<Rightarrow> ('b, 's) sm) \<Rightarrow> 'a list \<Rightarrow> ('b list, 's) sm"
where
  "mapm f xs = sequence (map f xs)"

lemma mapm_cong[fundef_cong]:
  assumes "xs = ys" and "\<And>x. x \<in> set ys \<Longrightarrow> f x = g x"
  shows "mapm f xs = mapm g ys"
proof -
  from assms have "map f xs = map g ys" by simp
  thus ?thesis unfolding mapm_def by simp
qed

lemmas monad_simps[simp] =
  return_bind bind_return bind_assoc
  run_simps

end
